//*******************************************************
//	FM channel Radio  mini Version
//	FM tuner module : NS953M (NIIGATA SEIMITSU CO., LTD.)
//
//	MCU:	ATmega88
//			efuse:F9h
//			hfuse:D6h	BOD 1.8V
//			lfuse:62h	内蔵CR 1MHz EEPROM保護
//
//	Compiler:	AVR-GCC		4.2.2
//				AVR-LIBC	1.4.7
//				binutils	2.18
//
//	2008.8.21	KAWAKAMI Yukio
//*******************************************************

#include	<avr/io.h>
#include	<avr/interrupt.h>
#include	<avr/pgmspace.h>
#include	<avr/eeprom.h>
#include	<util/twi.h>
#include	<string.h>


#define		CH_N		4

#define		SLEEP_MIN	30	// スリープモードで電源OFFする時間(分)

#include	"FMch.h"

// NS953M レジスタ初期値  R0 はクリアしておき、後で再設定する
PROGMEM uchar InitReg[] = {
	0x00, 0x00, 0xC0, 0x37,
	0xD8, 0x5D, 0x88, 0x4C,
	0x09, 0x4D, 0x28, 0x8C,
	0x04, 0x00 };

volatile uchar	Timer;
volatile uchar	FlashTimer;
volatile uchar	TwiTimer;
volatile uchar	FlashLED;
volatile uchar	FMch;
volatile uchar	SwFlag;
volatile uchar	LEDdata;
volatile uchar	SleepFlag;
volatile uchar	SleepTimerDS;
volatile uchar	SleepTimerS;
volatile uchar	SleepTimerM;

//*******************************************************
//	タイマー２比較Ａ比較一致割り込み
//	1MHz / 64 / 156 = 100Hz
//*******************************************************
ISR(TIMER2_COMPA_vect)
{
	uchar	d;

	// 各種タイマ
	if (Timer)	--Timer;
	if (FlashTimer)	--FlashTimer;
	if (TwiTimer)	--TwiTimer;
	if (SleepFlag){
		if (++SleepTimerDS >= 100){
			SleepTimerDS = 0;
			if (++SleepTimerS >= 60){
				SleepTimerS = 0;
				++SleepTimerM;
			}
		}
	}

	// スイッチ入力値を SwFlagにセット
	SwFlag = 0;
	d = ~PIND;

	if (d & (1<<P0))		SwFlag |= SW_1;
	if (d & (1<<P1))		SwFlag |= SW_2;
	if (d & (1<<POWER_SW))	SwFlag |= SW_POWER;
	if (d & (1<<P3))		SwFlag |= SW_3;
	if (d & (1<<P4))		SwFlag |= SW_4;

	// LED点滅処理
	if (FlashLED){
		if (FlashTimer == 0){
			LEDdata = 1 << FlashLED;
			FlashTimer = 50;
		} else if (FlashTimer == 25){
			LEDdata = 0;
		}
	}

	if (LEDdata & SW_1)		PORTD &= ~(1<<P5); else PORTD |= (1<<P5);
	if (LEDdata & SW_2)		PORTD &= ~(1<<P6); else PORTD |= (1<<P6);
	if (LEDdata & SW_3)		PORTD &= ~(1<<P7); else PORTD |= (1<<P7);
	if (LEDdata & SW_4)		PORTB &= ~(1<<P8); else PORTB |= (1<<P8);
}


//*******************************************************
//	メインルーチン
//*******************************************************
int main (void) {
	uint	freq;
	uchar	d;
	uchar	off_count;

	cli();	// 割り込み禁止

	// ポート設定
	PORTB = (1<<PB7)|(1<<PB6)|(1<<SCK)|(1<<MISO)|(1<<MOSI)|(1<<AMP_POWER)|(1<<P8)|(1<<P8);
	DDRB = (1<<AMP_POWER)|(1<<POWER_LOCK)|(1<<P8);

	PORTD = (1<<P0)|(1<<P1)|(1<<P2)|(1<<P3)|(1<<P4)|(1<<P5)|(1<<P6)|(1<<P7);
	DDRD = (1<<P5)|(1<<P6)|(1<<P7);

	PORTC = (1<<PC3)|(1<<PC2)|(1<<PC1)|(1<<PC0);
	DDRC = 0;

	// スリープ動作設定
	SMCR = (1<<SE);			// スリープ許可 アイドルモード

	// タイマ２ CTC動作
	TCCR2A = (1<<WGM21);
	TCCR2B = (1<<CS22);		// CTC動作 プリスケーラ 1/64
	OCR2A = 156-1;			// タイマカウント最大値 156  1MHz/64/156=100Hz
	TIMSK2 = (1<<OCIE2A);	// 比較Ａ割り込み許可

	// 電力カット
	PRR = (1<<PRTIM0)|(1<<PRTIM1)|(1<<PRSPI)|(1<<PRUSART0)|(1<<PRADC);

	ACSR = (1<<ADC);	// コンパレータ電源OFF

	sei();		// 割り込み許可

	// 約0.1秒時間待ち
	Timer = 10;
	while(Timer){
		WDR;
		SLEEP;
	}

	PORTB |= (1<<POWER_LOCK);	// 電源保持	

	// EEPROMから前回のチャンネル番号取得
	FMch = eeprom_read_byte((uint8_t *)EEPROM_CH);
	if (FMch > CH_N){
		FMch = 0;
	}
	LEDdata = 1 << (FMch + 1);	// LED点灯

	// NS953M リセット
	TWI_open();
	TWI_write(0xFF);
	TWI_start();
	TWI_close();

	// NS953M レジスタ初期値セット
	TWI_open();
	TWI_write(0xC8);	// スレーブアドレス + Write
	TWI_write(0);		// R0
	for (d = 0; d <= 13; d++){
		TWI_write(pgm_read_byte(&InitReg[d]));
	}
	TWI_close();

	// NS953M R0 セット
	TWI_open();
	TWI_write(0xC8);	// スレーブアドレス + Write
	TWI_write(0);		// R0
	if (eeprom_read_byte((uint8_t *)EEPROM_STEREO_FLAG) == 0xFF){
		TWI_write(0x11);	// PE=ON, SWSTD=1 Stereo
	} else {
		TWI_write(0x19);	// PE=ON, SWSTD=1 Mono
	}
	TWI_close();

	// SW2とSW4を押しながら電源を入れたら周波数設定モード
	if ((SwFlag & (SW_1|SW_2|SW_3|SW_4)) == (SW_2|SW_4)){
		scan_wave();	// 周波数スキャン＆設定
		FMch = 0;
	}

	LEDdata = 1 << (FMch + 1);
	freq = get_freq(FMch);	// EEPROMから周波数取得
	set_freq(freq);			// 周波数設定

	// スイッチを放すまで待つ
	while (SwFlag){
		SLEEP;
	}
	
	off_count = 0;
	for(;;){

		SLEEP;			// 10ms SLEEP

		if (SleepFlag && (SleepTimerM >= SLEEP_MIN)){
			power_off();	// 電源OFF
		}
		
		if (SwFlag & SW_POWER){

			// POWER SW
			if (++off_count >= 20){		// 200ms未満は無視
				LEDdata = 0;	// LED消灯
				// ボタンが放されるまで待つ
				while(SwFlag & SW_POWER){
					SLEEP;
					if (off_count < 255)	++off_count;
					if (off_count == 200){	// 2秒長押しでスリープモード
						FlashLED = FMch + 1;	// LED点滅
					}
				}
				if (off_count < 200){	// 2秒未満
					power_off();		// 電源OFF
				} else {
					// スリープモード設定
					SleepTimerDS = SleepTimerS = SleepTimerM = 0;
					SleepFlag = 1;
					FlashLED = 0;
					LEDdata = 1 << (FMch + 1);
				}
			}
		} else if (SwFlag){
			// 1〜4 SW
			FMch = 0;
			while(!(SwFlag & SW_1)){
				SwFlag >>= 1;
				FMch++;
			}
			LEDdata = 1 << (FMch + 1);			// LED点灯
			freq = get_freq(FMch);
			set_freq(freq);
			off_count = 0;

			// スイッチが放されるまで待つ
			while(SwFlag){
				SLEEP;
			}
		} else {
			off_count = 0;
		}
	}
}


//*******************************************************
//	電源OFF処理
//*******************************************************
void power_off(void){

	eeprom_write_byte((uint8_t *)EEPROM_CH, FMch);	// チャンネル番号保存
	cli();
	LEDdata = 0;
	PORTB &= ~(1<<POWER_LOCK);		// 電源保持を解除
	for (;;){		// 電源が切れる
		WDR;
		SLEEP;
	}
}


//*******************************************************
//	周波数スキャン＆設定
//*******************************************************
void scan_wave(void){
	uint	memory[CH_N];
	uchar	ch;
	uint	freq;
	uchar	sw;
	uchar	t;

	memset(memory, 0xFF,  sizeof(memory));
	
	// スイッチを放すまで LED1を点滅
	FlashLED = 1;
	while(SwFlag){
		SLEEP;
	}
	SLEEP;	// 念のため

	// 設定チャンネルボタン入力待ち
	t = 10;		// 10秒
	Timer = 100;
	while(!(sw = SwFlag)){
		SLEEP;
		if (!Timer){
			--t;
			if (!t){
				FlashLED = 0;	// 時間切れ
				return;
			} else {
				Timer = 100;
			}
		}
	}

	if (sw & SW_POWER){
		// POWER SWの場合は全チャンネルスキャン
		freq = 7600;
		ch = 0;
	} else {
		// チャンネルSWの場合はそのチャンネルデータを消去
		// 後のチャンネル周波数を詰める
		sw >>= 1;
		freq = 0xFFFF;
		for (ch = 0; ch < (CH_N - 1); ch++){
			sw >>= 1;
			if (sw){
				freq = get_freq(ch);
			} else {
				freq = get_freq(ch + 1);
			}
			memory[ch] = freq;
			if (freq == 0xFFFF)	break;
		}
		if (ch == (CH_N - 1)){
			freq = get_freq(ch);
		}
		if (freq == 0xFFFF)	freq = 7600 - 5;
		freq += 5;
	}
	FlashLED = ch + 1;

	// スイッチを放すまで待つ
	while(SwFlag){
		SLEEP;
	}
	SLEEP;	// 念のため
	
	// 周波数スキャン開始
	while(ch < CH_N){
		if (SwFlag & SW_POWER)	break;	// POWER SWで中断

		set_freq(freq);
		if (check_freq(freq)){	// 電波あり
			memory[ch++] = freq;
			FlashLED++;			// 次のLED点滅
		}
		if (freq >= 10800){
			break;
		} else {
			freq += 5;	// 50kHzステップ
		}
	}
	FlashLED = 0;

	if (!SwFlag){
		// 設定周波数書き込み
		for (ch = 0; ch < CH_N; ch++){
			eeprom_write_word((uint16_t *)(EEPROM_FREQ_START + (ch * 2)),
							  memory[ch]);
		}
	}
}


//*******************************************************
//	NS953Mの周波数関係のレジスタ設定
//*******************************************************
void set_freq(uint freq){
	ulong	data;
	uchar	r1, r2, r7;

	if ((freq < 7600)||(freq > 10800))	return;
	
	// R7 LOBIAS設定
	if (freq < 9400){			// < 94MHz
		r7 = 0x48 + 4;
	} else if (freq < 10000){	// < 100MHz
		r7 = 0x48 + 3;
	} else {
		r7 = 0x48 + 2;
	}

	// PLL設定値計算
	data = (ulong)freq * 10000L;	// 10kHz単位 → 1Hz単位
	data -= 375000L;				// IF 375kHz
	data /= 8192;					// PLLステップ 8192Hz

	r1 = (uchar)data;				// R1
	r2 = (uchar)(data >> 8) | 0xC0;	// R2

	TWI_open();			// NS953Mに設定
	TWI_write(0xC8);	// スレーブアドレス + Write
	TWI_write(1);		// R1
	TWI_write(r1);
	TWI_write(r2);
	TWI_close();

	TWI_open();
	TWI_write(0xC8);	// スレーブアドレス + Write
	TWI_write(7);		// R7
	TWI_write(r7);
	TWI_close();
}


//*******************************************************
//	チャンネル番号から周波数取得
//*******************************************************
uint get_freq(uchar ch){
	uint	freq;

	if (ch >= CH_N)	return 0xFFFF;

	// EEPROMから周波数値(10kHz単位)読み込み
	freq = eeprom_read_word((uint16_t *)(EEPROM_FREQ_START + (ch * 2)));

	return freq;
}


//*********************************************************
//	所定の周波数に電波があるか調べる
//*********************************************************
uchar check_freq(uint freq){
	uchar	r0;

	set_freq(freq);
	Timer = 12;
	r0 = read_NS953M(0);
	write_NS953M(0, (r0 & 0x0F) | 0x30);	// STO = RSS judgment
	while(Timer){ WDR; SLEEP; }
	if (!(read_NS953M(8) & 0x80))	return 0;	// なし
	write_NS953M(0, (r0 & 0x0F) | 0x20);	// STO = IF count judgment
	write_NS953M(0, (r0 & 0x0F) | 0x24);	// IFCST = 1
	Timer = 2;
	while(Timer){ WDR; SLEEP; }
	if (!(read_NS953M(8) & 0x80))	return 0;	// なし
	set_freq(freq + 5);
	write_NS953M(0, (r0 & 0x0F) | 0x60);	// STO = JL judgment
	Timer = 3;
	while(Timer){ WDR; SLEEP; }
	write_NS953M(0, (r0 & 0x0F) | 0x64);	// IFCST = 1
	Timer = 2;
	while(Timer){ WDR; SLEEP; }
	if (!(read_NS953M(8) & 0x80))	return 0;	// なし
	set_freq(freq - 5);
	write_NS953M(0, (r0 & 0x0F) | 0x50);	// STO = JU judgment
	Timer = 3;
	while(Timer){ WDR; SLEEP; }
	write_NS953M(0, (r0 & 0x0F) | 0x54);	// IFCST = 1
	Timer = 2;
	while(Timer){ WDR; SLEEP; }
	if (!(read_NS953M(8) & 0x80))	return 0;	// なし
	set_freq(freq);
	write_NS953M(0, r0);

	return 1;	// あった
}


//*********************************************************
//	NS953M レジスタに値をセット
//*********************************************************
void write_NS953M(uchar adrs, uchar data){

	TWI_open();
	TWI_write(0xC8);		// スレーブアドレス + Write
	TWI_write(adrs);		// レジスタアドレス
	TWI_write(data);
	TWI_close();
}


//*********************************************************
//	NS953M レジスタ値読み込み
//*********************************************************
uchar read_NS953M(uchar adrs){
	uchar data;

	TWI_open();
	TWI_write(0xC8);		// スレーブアドレス + Write
	TWI_write(adrs);		// レジスタアドレス
	TWI_start();
	TWI_write(0xC8 + 1);		// スレーブアドレス + Read
	data = TWI_read_nack();		// NACK
	TWI_close();

	return data;
}


//*******************************************************
// Clock module TWI Open
//*******************************************************
char TWI_open(void){

	TWBR = 0;
	TWSR = 0;		// TWI clock 62.5kHz

	TWCR = (1<<TWINT)|(1<<TWSTA)|(1<<TWEN);	// 開始条件送出
	TwiTimer = 10;	// 100ms
	while(TwiTimer){
		if (TWCR & (1<<TWINT))	break;
	}

	if ((TWSR & TW_NO_INFO) != TW_START){
		return 1;	// ERROR
	} else {
		return 0;
	}
}


//*******************************************************
//	TWI Write
//*******************************************************
char TWI_write(uchar data){

	TWDR = data;	// データ
	TWCR = (1<<TWINT)|(1<<TWEN);
	TwiTimer = 10;	// 100ms
	while(TwiTimer){
		if (TWCR & (1<<TWINT))	break;
	}

	if ((TWSR & TW_NO_INFO) != TW_MT_DATA_ACK){
		return 1;	// ERROR
	} else {
		return 0;
	}
}


//*******************************************************
// TWI Start
//*******************************************************
char TWI_start(void){

	TWCR = (1<<TWINT)|(1<<TWSTA)|(1<<TWEN);	// 開始条件送出

	TwiTimer = 10;	// 100ms
	while(TwiTimer){
		if (TWCR & (1<<TWINT))	break;
	}

	if ((TWSR & TW_NO_INFO) != TW_REP_START){
		return 1;	// ERROR
	} else {
		return 0;
	}
}


//*******************************************************
//	TWI Read ACK
//*******************************************************
uchar TWI_read_ack(void){

	TWCR = (1<<TWINT)|(1<<TWEN)|(1<<TWEA);

	TwiTimer = 10;	// 100ms
	while(TwiTimer){
		if (TWCR & (1<<TWINT))	break;
	}

	if ((TWSR & TW_NO_INFO) != TW_MR_DATA_ACK){
		return 0;		// ERROR
	} else {
		return TWDR;	// データ
	}
}


//*******************************************************
//	TWI Read NACK
//*******************************************************
uchar TWI_read_nack(void){

	TWCR = (1<<TWINT)|(1<<TWEN);

	TwiTimer = 10;	// 100ms
	while(TwiTimer){
		if (TWCR & (1<<TWINT))	break;
	}

	if ((TWSR & TW_NO_INFO) != TW_MR_DATA_NACK){
		return 0;		// ERROR
	} else {
		return TWDR;	// データ
	}
}


//*******************************************************
//	TWI Close
//*******************************************************
void TWI_close(void){

	TWCR = (1<<TWINT)|(1<<TWSTO)|(1<<TWEN);
}


